Index: if_iwi.c
===================================================================
RCS file: /cognet/ncvs/src/sys/dev/iwi/if_iwi.c,v
retrieving revision 1.39
diff -u -p -r1.39 if_iwi.c
--- if_iwi.c	26 Jul 2006 03:09:32 -0000	1.39
+++ if_iwi.c	12 Aug 2006 13:57:38 -0000
@@ -511,6 +511,12 @@ iwi_detach(device_t dev)
 	if (sc->mem != NULL)
 		bus_release_resource(dev, SYS_RES_MEMORY, sc->mem_rid, sc->mem);
 
+	if (sc->fw_map)
+		bus_dmamap_unload(sc->fw_dmat, sc->fw_map);
+	if (sc->fw_virtaddr != 0)
+		bus_dmamem_free(sc->fw_dmat, sc->fw_virtaddr, sc->fw_map);
+	if (sc->fw_dmat)
+		bus_dma_tag_destroy(sc->fw_dmat);
 	if (ifp != NULL)
 		if_free(ifp);
 
@@ -3060,6 +3066,7 @@ iwi_init_locked(void *priv, int force)
 	struct ifnet *ifp = ic->ic_ifp;
 	struct iwi_rx_data *data;
 	int i;
+	int must_realloc = 0;
 	IWI_LOCK_DECL;
 
 	if (sc->flags & IWI_FLAG_FW_LOADING)
@@ -3081,14 +3088,27 @@ iwi_init_locked(void *priv, int force)
 	}
 
 	/* allocate DMA memory for mapping firmware image */
-	if (sc->fw_boot.size > sc->fw_dma_size)
+	if (sc->fw_boot.size > sc->fw_dma_size) {
+		must_realloc = 1;
 		sc->fw_dma_size = sc->fw_boot.size;
-	if (sc->fw_fw.size > sc->fw_dma_size)
+	}
+	if (sc->fw_fw.size > sc->fw_dma_size) {
+		must_realloc = 1;
 		sc->fw_dma_size = sc->fw_fw.size;
-	if (sc->fw_uc.size > sc->fw_dma_size)
+	}
+	if (sc->fw_uc.size > sc->fw_dma_size) {
 		sc->fw_dma_size = sc->fw_uc.size;
+		must_realloc = 1;
+	}
+	if (must_realloc && sc->fw_virtaddr != 0) {
+		bus_dmamap_sync(sc->fw_dmat, sc->fw_map, BUS_DMASYNC_POSTWRITE);
+		bus_dmamap_unload(sc->fw_dmat, sc->fw_map);
+		bus_dmamem_free(sc->fw_dmat, sc->fw_virtaddr, sc->fw_map);
+		bus_dma_tag_destroy(sc->fw_dmat);
+	}
 
-	if (bus_dma_tag_create(NULL, 4, 0, BUS_SPACE_MAXADDR_32BIT,
+	if (must_realloc &&
+	    bus_dma_tag_create(NULL, 4, 0, BUS_SPACE_MAXADDR_32BIT,
 	    BUS_SPACE_MAXADDR, NULL, NULL, sc->fw_dma_size, 1, sc->fw_dma_size,
 	    0, NULL, NULL, &sc->fw_dmat) != 0) {
 		device_printf(sc->sc_dev,
@@ -3096,14 +3116,16 @@ iwi_init_locked(void *priv, int force)
 		IWI_LOCK(sc);
 		goto fail;
 	}
-	if (bus_dmamem_alloc(sc->fw_dmat, &sc->fw_virtaddr, 0,
+	if (must_realloc &&
+	    bus_dmamem_alloc(sc->fw_dmat, &sc->fw_virtaddr, 0,
 	    &sc->fw_map) != 0) {
 		device_printf(sc->sc_dev,
 		    "could not allocate firmware DMA memory\n");
 		IWI_LOCK(sc);
 		goto fail2;
 	}
-	if (bus_dmamap_load(sc->fw_dmat, sc->fw_map, sc->fw_virtaddr,
+	if (must_realloc &&
+	    bus_dmamap_load(sc->fw_dmat, sc->fw_map, sc->fw_virtaddr,
 	    sc->fw_dma_size, iwi_dma_map_addr, &sc->fw_physaddr, 0) != 0) {
 		device_printf(sc->sc_dev, "could not load firmware DMA map\n");
 		IWI_LOCK(sc);
@@ -3159,11 +3181,6 @@ iwi_init_locked(void *priv, int force)
 	}
 	sc->flags |= IWI_FLAG_FW_INITED;
 
-	bus_dmamap_sync(sc->fw_dmat, sc->fw_map, BUS_DMASYNC_POSTWRITE);
-	bus_dmamap_unload(sc->fw_dmat, sc->fw_map);
-	bus_dmamem_free(sc->fw_dmat, sc->fw_virtaddr, sc->fw_map);
-	bus_dma_tag_destroy(sc->fw_dmat);
-
 	if (iwi_config(sc) != 0) {
 		device_printf(sc->sc_dev, "device configuration failed\n");
 		goto fail;
@@ -3192,6 +3209,7 @@ fail4:	bus_dmamap_sync(sc->fw_dmat, sc->
 fail3:	bus_dmamem_free(sc->fw_dmat, sc->fw_virtaddr, sc->fw_map);
 fail2:	bus_dma_tag_destroy(sc->fw_dmat);
 fail:	ifp->if_flags &= ~IFF_UP;
+	sc->fw_virtaddr = 0;
 	sc->flags &= ~IWI_FLAG_FW_LOADING;
 	iwi_stop(sc);
 	iwi_put_firmware(sc);
